''
' floppyDriver.spin
'
 

CON    
 
  SECTOR_SIZE = 130    ' 128 longs plus 2 CRC plus a spare

  MOTOR_POWER  = 0     ' these 7 bits are parallel port outputs (hardcoded pin assignments) 
  DRIVE_ENABLE = 1
  HEAD_DIR     = 2
  HEAD_STEP    = 3
  HEAD_SELECT  = 4
  WRITE_ENABLE = 5
  WRITE_DATA   = 6
 
   
VAR

'------------------------------------
' floppy cog communication data structure

  long readyCounter               ' first cog writes counter here when new word ready
  long retrievedMFMwords          ' first cog writes new word here
  long requestedParameters        ' hub writes retrieval parms here for second cog
  long retrieveStatus             ' second cog writes status here for hub
  long sectorBuffer[SECTOR_SIZE]  ' second cog writes sector data here
'-------------------------------------

  long currentTrack               ' current floppy track setting 0-80

'--------------------------------------

PUB readFloppySector(sectorNumber)  | loops, indexPulse, track, side, sector

' tracks are numbered from 0 to 80; side 0,1; sectors from 1 to 18
' sectors start at track 0, side 0, sector 1
' first access all sectors, then the other side, then incr tracks.
' Some floppies have a track 81...

  if sectorNumber < 1 or sectorNumber > 2880
    return 1

  track  := (sectorNumber - 1) / 36
  side   := ((sectorNumber -1) / 18) & 1
  sector := ((sectorNumber - 1) // 18) + 1

' pack parameters into a long for the cog

  requestedParameters := (sector << 16) | (side << 8) | track

  drive_on
  setSide(side) 
  set_track(track)    

  startFloppy
  
  delay(500)
  
  retrieveStatus := HUBSAYSSTART
  
  repeat loops from 0 to 1000
    if retrieveStatus == validDataMask
      quit
    delay(2)
        
  motor_off
  drive_off
 
  
PUB startFloppy | indexPulse

' would like to start a separate cog to manage floppy motoring.
' Start motor, spin up. Include method to report when up to speed.
' Keep floppy running for several seconds after initial call
' in case subsequent calls also need to access it.

  motor_on  

' wait until the disk index bit shows 300 rpm
'  (about 1/5 second, 16,000,000 counts, between pulses)
' also implement a timeout in case of no disk, bad disk, etc.

  repeat
    indexPulse := ina[DISK_INDEX]    
  until indexPulse == 0

  delay(500)
    
  cognew(@byteReader, @readyCounter)
  cognew(@wordReader, @readyCounter)


PUB getStatus      
 
  return retrieveStatus

       
PUB getSectorArray(index)      
 
  return sectorBuffer[index]           


PUB delay(ms) | delt

  delt := (ms * 80000) #> 381    
  waitcnt(delt + cnt) 

      
PUB init | i  ' this contains hardcoded pin assignments
          
  currentTrack := -1
        
  repeat i from 0 to 7
    dira[i] := 1 
    outa[i] := 0

           
PUB set_track(togoto) | steps, trax, i

' set floppy head track. Return 0 on success; non-0 on failure
' Tested by reading sector header info from disk which holds track number 0-79(?)
  
  if currentTrack == -1
    if set_track_zero == 1
      return 1      
       
  if togoto < 0 or togoto > 80   ' by test, disk actually responds from 0 to 80; 81 total tracks!
    return 2

  steps := togoto 
   
  if togoto == currentTrack
    return 0
      
  if togoto < currentTrack
    steps := currentTrack - togoto - 1
    step_lower
    repeat i from 0 to steps  
      step_head
    currentTrack := togoto
    return 0 

  steps := togoto - currentTrack - 1
  step_higher
  repeat i from 0 to steps 
    step_head
      
  currentTrack := togoto
  return 0  
     
             
PUB set_track_zero | i  ' return 0 if track 1; 1 if failed to detect track zero 

  currentTrack := 0
  
  if ina[TRACK_ZERO] == 1
    return 0
        
  step_lower
  repeat i from 0 to 80
    step_head  
  
    if ina[TRACK_ZERO] == 1
      return 0

  return 1

  
PUB drive_on
  outa[DRIVE_ENABLE] := 1

  
PUB drive_off
  outa[DRIVE_ENABLE] := 0

  
PUB motor_on
  outa[MOTOR_POWER] := 1

      
PUB motor_off
  outa[MOTOR_POWER] := 0

    
PUB step_head
  outa[HEAD_STEP] := 1   
  waitcnt(400000 + cnt)   ' 400,000 seems ok
  outa[HEAD_STEP] := 0    
  waitcnt(400000 + cnt)

    
PUB step_higher
  outa[HEAD_DIR] := 1

  
PUB step_lower
  outa[HEAD_DIR] := 0

  
PUB setSide(head)
  if head == 0
    outa[HEAD_SELECT] := 0
  
  if head == 1
    outa[HEAD_SELECT] := 1


CON


  READ_DATA    = 13     ' these are three parallel port inputs
  TRACK_ZERO   = 14
  DISK_INDEX   = 15

       
DAT

org

' Read pulse stream from floppy read head.
' Detect unique sync pulses to bit-align the 16-bit MFM output words.
' After each additional 16 bits have accumulated, send the entire 32-bit
' current-word plus previous-word to the hub, followd by the current counter value.

byteReader                         

        mov     cog1CounterAddr , par 
        mov     cog1DataAddr, par
        add     cog1DataAddr, #4
               
        mov     bitcount, #16                                         
        shl     fluxBit, READ_DATA_BIT
        mov     start, cnt
                         
:top

' wait for the next magnetic flux transition pulse

        waitpne fluxBit, fluxBit
        waitpeq fluxBit, fluxBit            ' capture time AFTER the PEQ transition...
        mov     afterpne, cnt          

' calculate clock counts since last pulse
         
        mov     pulsecount, afterpne
        sub     pulsecount, start
        mov     start, afterpne

'-------------------------
' Add 2-4 bits to the raw MFM word based on latest pulse interval.
' warning - there are about 160 clocks until the next pulse may be detected,
' so the longest code path should use less than 40 instructions.
' 
' First bit - mandatory one.

        shl     rawbits, #1                  ' place a "1" bit into the raw MFM word.
        or      rawbits, #1

' test for 32-bit "sync-sync" combination  (which ends in a "1" bit) 
                                                                          
        cmp     rawBits, syncSync wc, wz  
  IF_NE jmp     #:notSync1            ' 
        mov     bitcount, #1                ' align bit position

:notSync1
        djnz    bitcount,    #:nextBit1     ' skip byte send until 16 bits seen
        wrLong  rawbits, cog1DataAddr
        mov     counter, cnt
        mov     bitcount, #16
        wrLong  counter, cog1CounterAddr

'-------------------------
' second bit - mandatory zero

:nextBit1

        shl     rawbits, #1                 ' Add the mandatory zero bit.                                             
        djnz    bitcount,    #:nextBit2     ' skip byte send until 16 bits seen
        wrLong  rawbits, cog1DataAddr
        mov     counter, cnt
        mov     bitcount, #16
        wrLong  counter, cog1CounterAddr

'-------------------------
' third bit - occasionally add another zero. (If we do not add another zero
' here, then presumably some time is available, so recalibrate the pulsecount
' transition point based on the current value of the delay.) 

:nextBit2

        cmp     pulsecount, D1 wc, wz       ' If pulselength is short then 
  IF_B  jmp     #:recalibrate               ' do not add another zero bit.  
        shl     rawbits, #1 
        djnz    bitcount,    #:nextBit3     ' skip byte send until 16 bits seen
        wrLong  rawbits, cog1DataAddr
        mov     counter, cnt
        mov     bitcount, #16
        wrLong  counter, cog1CounterAddr

'-------------------------
' fourth bit - occasionally add another zero

:nextBit3
        cmp     pulsecount, D2 wc, wz     ' If pulselength is short then 
  IF_B  jmp     #:top                       ' do not add another zero bit.              
        shl     rawbits, #1
        djnz    bitcount,    #:top          ' skip byte send until 16 bits seen
        wrLong  rawbits, cog1DataAddr
        mov     counter, cnt
        mov     bitcount, #16
        wrLong  counter, cog1CounterAddr
        jmp     #:top
 
'---------------------------------------
' adjust the timing interval.

:recalibrate

' basic unit of delay is about 80.
' pulses are separated by 2, 3 or 4 of those units, for 160, 240 or 320 counts.
' So if another pulse arrives when the count is less than 200, only append one
' zero bit. If less than 280, append two zero bits. Otherwise append three zero bits.
'
' Maintain half a unit around 40.
' 5 times 40, 200, is D1, the first decision value,
' 7 times 40, 280, is D2, the second decision value.
' 
' right now, pulsecount should be near 160. No fancy averaging or outlier detection
' and elimination; just increment or decrement the basic unit every pass.

        shr     pulsecount, #2
        cmp     pulsecount, halfunit wc, wz
  IF_E  jmp     #:top
  IF_B  add     halfunit, #1
  IF_A  sub     halfunit, #1

        mov     halfunit, #40
        
        mov     D1, halfunit
        shl     D1, #2          ' (40 * 4) + 40 = 200
        add     D1, halfunit
        mov     D2, halfunit
        shl     D2, #3           ' (40 * 8) - 40 = 280
        sub     D2, halfunit
    
        jmp     #:top

halfunit   long 40
D1         long 200
D2         long 280
        
'---------------------------------------
' syncsync is a unique bit pattern. It is a legal RLL(1,3) pattern, but is
' a pattern that is guaranteed to never appear in any normal data.  

syncSync             long %0100_0100_1000_1001_0100_0100_1000_1001
cog1CounterAddr      long 0
cog1MfmWordCount     long 0
cog1DataAddr         long 0
bitcount             long 0
rawbits              long 0 
fluxBit              long 1
READ_DATA_BIT        long READ_DATA
afterpne             long 0
start                long 0
pulsecount           long 0
counter              long 0


'*************************************************************************

CON 





DAT

org

' read hub MFM words every time the value in the MFMCount changes.
' Decode, identify header and body data, write to hub array.
' Set progress flags every step of the way, so if this fails to locate a sector
' we have a clue what failed.

wordReader                         

        mov     cog2CounterAddr, par 
        mov     dataByteAddr, par
        add     dataByteAddr, #4 
        
        mov     requestedParametersAddr, par
        add     requestedParametersAddr, #8
       
        mov     asmRetrieveStatusAddr, par
        add     asmRetrieveStatusAddr, #12
                     
        mov     dataArrayBase , par
        add     dataArrayBase , #16        
        mov     dataArrayPtr, dataArrayBase         

                      
' Two reasons to be waiting at spinwait
' 1. have not started to do anything     (validFlag is 0)
' 2. finished a sector retrieval request (validFlag holds some statuses)     
        
:spinwait
       rdLong  validFlag, asmRetrieveStatusAddr
       tjz     validFlag, #:spinwait
       and     validFlag, RETRIEVALCOMPLETED wz, wc, nr
 IF_NZ jmp     #:spinwait       


' begin a sector retrieval pass

        mov    totalBytes, #0

'---------------------------------------
' Wait for an incoming MFM word.
                        
:wait            
        rdLong  sequenceCount, cog2CounterAddr
        cmp     sequenceCount, prevsequenceCount wc, wz
 IF_E   jmp     #:wait        
        mov     prevsequenceCount, sequenceCount

        rdLong  mfmWord, dataByteAddr

'---------------------------------
' timeout check
' only allow 10000 or so bytes to pass before declaring timeout.

        add     totalBytes, #1
        cmp     totalBytes, TRACKTIMEOUTCOUNT wc, wz
 IF_B   jmp     #:tryHeader
        or      validFlag, TRACKTIMEOUT
        or      validFlag, RETRIEVALCOMPLETED
        wrLong  validFlag, asmRetrieveStatusAddr    
        jmp     #:spinwait
            
'-----------------------------------
' check for the special section id

:tryHeader
        cmp     mfmWord, syncHeader wc, wz
 IF_NE  jmp     #:tryBody
        mov     state, HEADER
        mov     headerCount, #0
        mov     crc, headerCrcInit

        rdLong  asmParameters, requestedParametersAddr      
        mov     requestedTrack, asmParameters
        and     requestedTrack, #$FF 
        mov     requestedSide,  asmParameters
        shr     requestedSide, #8
        and     requestedSide, #$FF        
        mov     requestedSector, asmParameters
        shr     requestedSector, #16
        and     requestedSector, #$FF
 
        or      validFlag, HDRIDSEEN
        wrLong  validFlag, asmRetrieveStatusAddr       
        jmp     #:wait

' if the current word is not the header, then test for the data body id mark

:tryBody 
        cmp     mfmWord, syncBody wc, wz        
 IF_NE  jmp     #:regularByte
        mov     state, BODY
        mov     bodyCount, #0
        mov     crc, bodyCrcInit
        mov     dataArrayPtr, dataArrayBase         
        mov     pushIndex, #1
        or      validFlag, BODYIDSEEN
        wrLong  validFlag, asmRetrieveStatusAddr        
        jmp     #:wait

' otherwise treat as a normal data word
 
:regularByte

        call    #crc_mfm
        cmp     state, HEADER wc, wz
  IF_E  jmp     #:processHeader
        cmp     state, BODY wc, wz
  IF_E  jmp     #:processBody
        jmp     #:wait
         
'-----------------------------------

:processHeader        'bytes 1-6: 4 data + 2 CRC

        add     headerCount, #1
        call    #oneByteCRC
                                 
        cmp     headerCount, #1 wc, wz
 IF_A   jmp     #:testByte2        
        cmp     dataByte, requestedTrack wc, wz
 IF_NE  jmp     #:resetHeader 
        or      validFlag, HDRTRACKVALID
        wrLong  validFlag, asmRetrieveStatusAddr
        jmp     #:wait
                              
:testByte2
        cmp     headerCount, #2 wc, wz
 IF_A   jmp     #:testByte3
        cmp     dataByte, requestedSide wc, wz
 IF_NE  jmp     #:resetHeader
        or      validFlag, HDRSIDEVALID
        wrLong  validFlag, asmRetrieveStatusAddr
        jmp     #:wait        

:testByte3    
        cmp     headerCount, #3 wc, wz
 IF_A   jmp     #:testByte4
        cmp     dataByte, requestedSector wc, wz
 IF_NE  jmp     #:resetHeader
        or      validFlag, HDRSECTORVALID
        wrLong  validFlag, asmRetrieveStatusAddr        
        jmp     #:wait
        
:testByte4
        cmp     headerCount, #4 wc, wz
 IF_A   jmp     #:testByte6
        cmp     dataByte, #2 wc, wz       ' Always expect the media code to be "2"
 IF_NE  jmp     #:resetHeader 
        or      validFlag, HDRMEDIAVALID
        wrLong  validFlag, asmRetrieveStatusAddr                
        jmp     #:wait
        
:testByte6
        cmp     headerCount, #6 wc, wz
 IF_A   jmp     #:testByte8
        cmp     crc, #0 wc, wz
 IF_NE  jmp     #:resetHeader 
        or      validFlag, HDRCRCTESTVALID
        wrLong  validFlag, asmRetrieveStatusAddr
        jmp     #:wait  

:testByte8
        cmp     headerCount, #8 wc, wz
 IF_AE  mov     state, NULL
        jmp     #:wait

:resetHeader
        mov     state, NULL
        jmp     #:wait 
 
'-------------------------------------
' process body data

:processBody  
  
        add     bodyCount, #1
        call    #oneByteCRC
                        
        cmp     validFlag, validSaveMask wc, wz
 IF_NE  jmp     #:resetBody

        wrByte  dataByte, dataArrayPtr       
        add     dataArrayPtr, #1
        add     pushIndex, #1

' At byte 514, check the results of the data CRC
          
        cmp     bodyCount, five14 wc, wz
 IF_B   jmp     #:wait       
 IF_A   jmp     #:resetBody
 
        cmp     crc, #0 wc, wz
 IF_NE  jmp     #:resetBody 
        or      validFlag, DATACRCTESTVALID
        or      validFlag, RETRIEVALCOMPLETED
        wrLong  validFlag, asmRetrieveStatusAddr        
        jmp     #:resetBody

:resetBody
        mov     state, NULL        
        jmp     #:spinwait 

'-------------------------------------------------

oneByteCRC

        mov     loop, #8
        mov     mask, #128
:crcloop
        mov     xor_flag, crc
        and     crc, bit16
        shl     crc, #1
        and     dataByte, mask wc, wz, nr
 IF_NZ  add     crc, #1
        shr     mask, #1
        cmp     xor_flag, #0 wc, wz 
 IF_NZ  xor     crc, CCITT_POLYNOMIAL
        djnz    loop, #:crcloop           
        and     crc, bit17

oneByteCRC_ret ret
         
' The following routine is supposed to both translate the 16-bit MFM word to an 8 bit
' data byte and adjust the current CRC state for this data set, but the CRC part doesn't
' work right. The CRC code above does work, tho. Maybe someday I'll fix the crc_mfm code.
 
crc_mfm

        mov      dataByte, #0
        
' pass 7
        and     asmCrc,   bit14 wc, nr     ' set carry if bit14 of CRC is set but don't change value of asmCrc.
        shl     asmCrc,   #1               ' Every odd bit of 16-bit MFM word is the binary bit value,
        and     mfmWord,  bit15 wz, nr     ' so set the zero flag based on every odd bit.    
  IF_NZ or      dataByte, bit8             ' A not-set Zero flag means the data bit was set, so both       
  IF_NZ add     asmCrc,   #1               ' build the result and process the CRC based on that flag.
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL ' Lastly continue with CRC value processing.
' pass 6
        and     asmCrc,   bit14 wc, nr
        shl     asmCrc,   #1
        and     mfmWord,  bit13 wz, nr
  IF_NZ or      dataByte, bit7  
  IF_NZ add     asmCrc,   #1
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL
' pass 5
        and     asmCrc,   bit14 wc, nr
        shl     asmCrc,   #1
        and     mfmWord,  bit11 wz, nr
  IF_NZ or      dataByte, bit6 
  IF_NZ add     asmCrc,   #1
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL
' pass 4
        and     asmCrc,   bit14 wc, nr 
        shl     asmCrc,   #1
        and     mfmWord,  bit9 wz, nr
  IF_NZ or      dataByte, bit5  
  IF_NZ add     asmCrc,   #1
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL 
' pass 3
        and     asmCrc,   bit14 wc, nr 
        shl     asmCrc,   #1
        and     mfmWord,  bit7 wz, nr
  IF_NZ or      dataByte, bit4 
  IF_NZ add     asmCrc,   #1
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL
' pass 2
        and     asmCrc,   bit14 wc, nr
        shl     asmCrc,   #1
        and     mfmWord,  bit5 wz, nr
  IF_NZ or      dataByte, bit3 
  IF_NZ add     asmCrc,   #1
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL
' pass 1
        and     asmCrc,   bit14 wc, nr
        shl     asmCrc,   #1
        and     mfmWord,  bit3 wz, nr
  IF_NZ or      dataByte, bit2 
  IF_NZ add     asmCrc,   #1
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL
' pass 0
        and     asmCrc,   bit14 wc, nr 
        shl     asmCrc,   #1
        and     mfmWord,  bit1 wz, nr
  IF_NZ or      dataByte, bit1  
  IF_NZ add     asmCrc,   #1
  IF_C  xor     asmCrc,   CCITT_POLYNOMIAL

' done. dataByte contains the decoded MFM value, and asmCrc contains the current CRC value.
' asmCrc will be zero at the end of a group of values if no data error occurred.

crc_mfm_ret ret

loop     long 0
mask     long 128
xor_flag long 0
crc      long 0

totalBytes        long 0
TRACKTIMEOUTCOUNT long 10000

'-------------------------------------------------
syncHeader  long $4489_5554
syncBody    long $4489_5545
headerCount long 0
bodyCount   long 0
mfmWord     long 0

asmParameters long 0


bit1                 long %0000_0000_0000_0001   '     1
bit2                 long %0000_0000_0000_0010   '     2
bit3                 long %0000_0000_0000_0100   '     4
bit4                 long %0000_0000_0000_1000   '     8
bit5                 long %0000_0000_0001_0000   '    16
bit6                 long %0000_0000_0010_0000   '    32
bit7                 long %0000_0000_0100_0000   '    64
bit8                 long %0000_0000_1000_0000   '   128
bit9                 long %0000_0001_0000_0000   '   256
bit10                long %0000_0010_0000_0000   '   512
bit11                long %0000_0100_0000_0000   '  1024
bit12                long %0000_1000_0000_0000   '  2048
bit13                long %0001_0000_0000_0000   '  4096
bit14                long %0010_0000_0000_0000   '  8192
bit15                long %0100_0000_0000_0000   ' 16384
bit16                long %1000_0000_0000_0000   ' 32768
bit17                long $0001_0000             ' 65536

'-------------------------------------------------

pushIndex    long 0 
lastData     long 0
  
cog2CounterAddr      long 0                              
dataByteAddr         long 0
dataByte             long 0
asmProcessFlagAddr   long 0 
asmProcessFlag       long 0

requestedParametersAddr  long 0
parameterPtr             long 0

desiredTrack          long 0
desiredSide           long 0
desiredSector         long 0

asmRetrieveStatusAddr long 0
validFlag             long 0

dataArrayBase        long 0
dataArrayPtr         long 0
dataArraySize        long SECTOR_SIZE 
byteArraySize        long SECTOR_SIZE * 4
sequenceCount        long 0
prevSequenceCount    long 0

asmCrc               long 0
state                long 0

requestedTrack       long 0
requestedSide        long 0
requestedSector      long 0
requestedMedia       long 2

CCITT_POLYNOMIAL     long 4129
five14               long 514

headerCrcInit        long 14197      ' These are "magic" numbers that initialize the CRC for each part.
bodyCrcInit          long 14192      ' They work, but I have no idea why it is done like this. 

BODY                 long 1
HEADER               long 2
NULL                 long 3

' ValidFlag Bits 
'
' Each test sets a bit. Any failed test invalidates the entire sector retrieval.
' There is an intermediate save mask after the header tests OK, although the data
' CRC has not yet been performed.

HUBSAYSSTART           long %1
HDRIDSEEN              long %10              
BODYIDSEEN             long %100          
HDRTRACKVALID          long %1000         ' F

HDRSIDEVALID           long %10000 
HDRSECTORVALID         long %100000       
HDRMEDIAVALID          long %1000000      
HDRCRCTESTVALID        long %10000000     ' F

validSaveMask          long %11111111     ' allow to write data to array when these many bits set
 
DATACRCTESTVALID       long %100000000
RETRIEVALCOMPLETED     long %1000000000

validDataMask          long %1111111111  ' but final check isn't complete until data CRC passes.

TRACKTIMEOUT           long %10000000000

'----------------------------------------------